#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
FreeType high-level python API

Adapted from freetype-py.
'''

import sys
import struct
from ctypes import (byref, c_char_p, c_ushort, cast, util, CDLL, Structure,
                    POINTER, c_int, c_short, c_long, c_void_p, c_uint,
                    c_char, c_ubyte, CFUNCTYPE)

from ..six import string_types
from ...util import load_data_file

FT_LOAD_RENDER = 4
FT_KERNING_DEFAULT = 0
FT_KERNING_UNFITTED = 1
FT_LOAD_NO_HINTING = 2
FT_LOAD_FORCE_AUTOHINT = 32
FT_LOAD_NO_AUTOHINT = 32768
FT_LOAD_TARGET_LCD = 196608
FT_LOAD_TARGET_LIGHT = 65536
FT_LOAD_NO_SCALE = 1
FT_FACE_FLAG_SCALABLE = 1

_64_bit = (8 * struct.calcsize("P")) == 64
##############################################################################
# ft_structs

FT_Int = c_int
FT_UInt = c_uint
FT_F2Dot14 = c_short
FT_Pos = FT_Fixed = FT_Long = c_long
FT_Glyph_Format = c_int
FT_String_p = c_char_p
FT_Short = c_short  # A typedef for signed short.
FT_UShort = c_ushort  # A typedef for unsigned short.
FT_Generic_Finalizer = CFUNCTYPE(None, c_void_p)
FT_Encoding = c_int


class FT_LibraryRec(Structure):
    _fields_ = []
FT_Library = POINTER(FT_LibraryRec)


class FT_Vector(Structure):
    _fields_ = [('x', FT_Pos), ('y', FT_Pos)]


class FT_UnitVector(Structure):
    _fields_ = [('x', FT_F2Dot14), ('y', FT_F2Dot14)]


class FT_Matrix(Structure):
    _fields_ = [('xx', FT_Fixed), ('xy', FT_Fixed),
                ('yx', FT_Fixed), ('yy', FT_Fixed)]


class FT_GlyphRec(Structure):
    _fields_ = [('library', FT_Library), ('clazz', c_void_p),
                ('format', FT_Glyph_Format), ('advance', FT_Vector)]
FT_Glyph = POINTER(FT_GlyphRec)


class FT_Bitmap(Structure):
    _fields_ = [('rows', c_int), ('width', c_int),
                ('pitch', c_int), ('buffer', POINTER(c_ubyte)),
                ('num_grays', c_short), ('pixel_mode', c_ubyte),
                ('palette_mode', c_char), ('palette', c_void_p)]


class FT_BitmapGlyphRec(Structure):
    _fields_ = [('root', FT_GlyphRec), ('left', FT_Int),
                ('top', FT_Int), ('bitmap', FT_Bitmap)]
FT_BitmapGlyph = POINTER(FT_BitmapGlyphRec)


class FT_Glyph_Metrics(Structure):
    _fields_ = [('width', FT_Pos), ('height', FT_Pos),
                ('horiBearingX', FT_Pos), ('horiBearingY', FT_Pos),
                ('horiAdvance',  FT_Pos), ('vertBearingX', FT_Pos),
                ('vertBearingY', FT_Pos), ('vertAdvance',  FT_Pos)]


class FT_Outline(Structure):
    _fields_ = [('n_contours', c_short), ('n_points', c_short),
                ('points', POINTER(FT_Vector)), ('tags', POINTER(c_ubyte)),
                ('contours', POINTER(c_short)), ('flags', c_int)]


class FT_Size_Metrics(Structure):
    _fields_ = [('x_ppem', FT_UShort), ('y_ppem', FT_UShort),
                ('x_scale', FT_Fixed), ('y_scale', FT_Fixed),
                ('ascender', FT_Pos),  ('descender', FT_Pos),
                ('height', FT_Pos), ('max_advance', FT_Pos)]


class FT_BBox(Structure):
    _fields_ = [('xMin', FT_Pos), ('yMin', FT_Pos),
                ('xMax', FT_Pos), ('yMax', FT_Pos)]


class FT_Generic(Structure):
    _fields_ = [('data', c_void_p), ('finalizer', FT_Generic_Finalizer)]


class FT_SizeRec(Structure):
    _fields_ = [('face', c_void_p), ('generic', FT_Generic),
                ('metrics', FT_Size_Metrics), ('internal', c_void_p)]
FT_Size = POINTER(FT_SizeRec)


class FT_CharmapRec(Structure):
    _fields_ = [('face', c_void_p), ('encoding', FT_Encoding),
                ('platform_id', FT_UShort), ('encoding_id', FT_UShort)]
FT_Charmap = POINTER(FT_CharmapRec)


class FT_Bitmap_Size(Structure):
    _fields_ = [('height', FT_Short), ('width', FT_Short),
                ('size', FT_Pos), ('x_ppem', FT_Pos), ('y_ppem', FT_Pos)]


class FT_GlyphSlotRec(Structure):
    _fields_ = [('library', FT_Library), ('face', c_void_p),
                ('next', c_void_p), ('reserved', c_uint),
                ('generic', FT_Generic), ('metrics', FT_Glyph_Metrics),
                ('linearHoriAdvance', FT_Fixed),
                ('linearVertAdvance', FT_Fixed),
                ('advance', FT_Vector), ('format',  FT_Glyph_Format),
                ('bitmap', FT_Bitmap), ('bitmap_left', FT_Int),
                ('bitmap_top', FT_Int), ('outline', FT_Outline),
                ('num_subglyphs', FT_UInt), ('subglyphs', c_void_p),
                ('control_data', c_void_p), ('control_len', c_long),
                ('lsb_delta', FT_Pos), ('rsb_delta', FT_Pos),
                ('other', c_void_p), ('internal', c_void_p)]
FT_GlyphSlot = POINTER(FT_GlyphSlotRec)


class FT_FaceRec(Structure):
    _fields_ = [('num_faces', FT_Long), ('face_index', FT_Long),
                ('face_flags', FT_Long), ('style_flags', FT_Long),
                ('num_glyphs', FT_Long), ('family_name', FT_String_p),
                ('style_name', FT_String_p), ('num_fixed_sizes', FT_Int),
                ('available_sizes', POINTER(FT_Bitmap_Size)),
                ('num_charmaps', c_int), ('charmaps', POINTER(FT_Charmap)),
                ('generic', FT_Generic), ('bbox', FT_BBox),
                ('units_per_EM', FT_UShort), ('ascender', FT_Short),
                ('descender', FT_Short), ('height', FT_Short),
                ('max_advance_width', FT_Short),
                ('max_advance_height', FT_Short),
                ('underline_position',  FT_Short),
                ('underline_thickness', FT_Short),
                ('glyph', FT_GlyphSlot), ('size', FT_Size),
                ('charmap', FT_Charmap),
                ('driver', c_void_p), ('memory', c_void_p),
                ('stream', c_void_p), ('sizes_list_head', c_void_p),
                ('sizes_list_tail', c_void_p), ('autohint', FT_Generic),
                ('extensions', c_void_p), ('internal', c_void_p)]
FT_Face = POINTER(FT_FaceRec)


##############################################################################
# __init__.py

__dll__ = None
FT_Library_filename = util.find_library('freetype')
if not FT_Library_filename and sys.platform.startswith('win'):
    fname_end = '_x64.dll' if _64_bit else '.dll'
    FT_Library_filename = load_data_file('freetype/freetype253' + fname_end)
if not FT_Library_filename:
    raise ImportError('Freetype library not found')
if not __dll__:
    __dll__ = CDLL(FT_Library_filename)


FT_Init_FreeType = __dll__.FT_Init_FreeType
FT_Done_FreeType = __dll__.FT_Done_FreeType
FT_Library_Version = __dll__.FT_Library_Version
__handle__ = None


# Comment out to avoid segfaults on Py34
# def __del_library__(self):
#     global __handle__
#     if __handle__:
#         try:
#             FT_Done_FreeType(self)
#             __handle__ = None
#         except:
#             pass
# FT_Library.__del__ = __del_library__


def get_handle():
    '''
    Get unique FT_Library handle
    '''
    global __handle__
    if not __handle__:
        __handle__ = FT_Library()
        error = FT_Init_FreeType(byref(__handle__))
        if error:
            raise RuntimeError(hex(error))
    return __handle__


def version():
    '''
    Return the version of the FreeType library being used as a tuple of
    ( major version number, minor version number, patch version number )
    '''
    amajor = FT_Int()
    aminor = FT_Int()
    apatch = FT_Int()
    library = get_handle()
    FT_Library_Version(library, byref(amajor), byref(aminor), byref(apatch))
    return (amajor.value, aminor.value, apatch.value)


try:
    FT_Library_SetLcdFilter = __dll__.FT_Library_SetLcdFilter
except:
    def FT_Library_SetLcdFilter(*args, **kwargs):
        return 0
if version() >= (2, 4, 0):
    FT_Library_SetLcdFilterWeights = __dll__.FT_Library_SetLcdFilterWeights
FT_New_Face = __dll__.FT_New_Face
FT_New_Memory_Face = __dll__.FT_New_Memory_Face
FT_Open_Face = __dll__.FT_Open_Face
FT_Attach_File = __dll__.FT_Attach_File
FT_Attach_Stream = __dll__.FT_Attach_Stream
if version() >= (2, 4, 2):
    FT_Reference_Face = __dll__.FT_Reference_Face
FT_Done_Face = __dll__.FT_Done_Face
FT_Done_Glyph = __dll__.FT_Done_Glyph
FT_Select_Size = __dll__.FT_Select_Size
FT_Request_Size = __dll__.FT_Request_Size
FT_Set_Char_Size = __dll__.FT_Set_Char_Size
FT_Set_Pixel_Sizes = __dll__.FT_Set_Pixel_Sizes
FT_Load_Glyph = __dll__.FT_Load_Glyph
FT_Load_Char = __dll__.FT_Load_Char
FT_Set_Transform = __dll__.FT_Set_Transform
FT_Render_Glyph = __dll__.FT_Render_Glyph
FT_Get_Kerning = __dll__.FT_Get_Kerning
FT_Get_Track_Kerning = __dll__.FT_Get_Track_Kerning
FT_Get_Glyph_Name = __dll__.FT_Get_Glyph_Name
FT_Get_Glyph = __dll__.FT_Get_Glyph

FT_Glyph_Get_CBox = __dll__.FT_Glyph_Get_CBox

FT_Get_Postscript_Name = __dll__.FT_Get_Postscript_Name
FT_Get_Postscript_Name.restype = c_char_p
FT_Select_Charmap = __dll__.FT_Select_Charmap
FT_Set_Charmap = __dll__.FT_Set_Charmap
FT_Get_Charmap_Index = __dll__.FT_Get_Charmap_Index
FT_Get_CMap_Language_ID = __dll__.FT_Get_CMap_Language_ID
FT_Get_CMap_Format = __dll__.FT_Get_CMap_Format
FT_Get_Char_Index = __dll__.FT_Get_Char_Index
FT_Get_First_Char = __dll__.FT_Get_First_Char
FT_Get_Next_Char = __dll__.FT_Get_Next_Char
FT_Get_Name_Index = __dll__.FT_Get_Name_Index
FT_Get_SubGlyph_Info = __dll__.FT_Get_SubGlyph_Info
if version() >= (2, 3, 8):
    FT_Get_FSType_Flags = __dll__.FT_Get_FSType_Flags
    FT_Get_FSType_Flags.restype = c_ushort

FT_Get_X11_Font_Format = __dll__.FT_Get_X11_Font_Format
FT_Get_X11_Font_Format.restype = c_char_p

FT_Get_Sfnt_Name_Count = __dll__.FT_Get_Sfnt_Name_Count
FT_Get_Sfnt_Name = __dll__.FT_Get_Sfnt_Name
FT_Get_Advance = __dll__.FT_Get_Advance


FT_Outline_GetInsideBorder = __dll__.FT_Outline_GetInsideBorder
FT_Outline_GetOutsideBorder = __dll__.FT_Outline_GetOutsideBorder
FT_Outline_Get_BBox = __dll__.FT_Outline_Get_BBox
FT_Outline_Get_CBox = __dll__.FT_Outline_Get_CBox
FT_Stroker_New = __dll__.FT_Stroker_New
FT_Stroker_Set = __dll__.FT_Stroker_Set
FT_Stroker_Rewind = __dll__.FT_Stroker_Rewind
FT_Stroker_ParseOutline = __dll__.FT_Stroker_ParseOutline
FT_Stroker_BeginSubPath = __dll__.FT_Stroker_BeginSubPath
FT_Stroker_EndSubPath = __dll__.FT_Stroker_EndSubPath
FT_Stroker_LineTo = __dll__.FT_Stroker_LineTo
FT_Stroker_ConicTo = __dll__.FT_Stroker_ConicTo
FT_Stroker_CubicTo = __dll__.FT_Stroker_CubicTo
FT_Stroker_GetBorderCounts = __dll__.FT_Stroker_GetBorderCounts
FT_Stroker_ExportBorder = __dll__.FT_Stroker_ExportBorder
FT_Stroker_GetCounts = __dll__.FT_Stroker_GetCounts
FT_Stroker_Export = __dll__.FT_Stroker_Export
FT_Stroker_Done = __dll__.FT_Stroker_Done
FT_Glyph_Stroke = __dll__.FT_Glyph_Stroke
FT_Glyph_StrokeBorder = __dll__.FT_Glyph_StrokeBorder
FT_Glyph_To_Bitmap = __dll__.FT_Glyph_To_Bitmap


Vector = FT_Vector
Matrix = FT_Matrix


class Bitmap(object):
    def __init__(self, bitmap):
        self._FT_Bitmap = bitmap

    rows = property(lambda self: self._FT_Bitmap.rows)
    width = property(lambda self: self._FT_Bitmap.width)
    pitch = property(lambda self: self._FT_Bitmap.pitch)
    buffer = property(lambda self:
                      [self._FT_Bitmap.buffer[i]
                       for i in range(self.rows * self.pitch)])


class Glyph(object):
    def __init__(self, glyph):
        self._FT_Glyph = glyph

    def __del__(self):
        if self._FT_Glyph is not None and FT_Done_Glyph is not None:
            FT_Done_Glyph(self._FT_Glyph)

    def to_bitmap(self, mode, origin, destroy=False):
        error = FT_Glyph_To_Bitmap(byref(self._FT_Glyph),
                                   mode, origin, destroy)
        if error:
            raise RuntimeError(hex(error))
        return BitmapGlyph(self._FT_Glyph)


class BitmapGlyph(object):
    def __init__(self, glyph):
        self._FT_BitmapGlyph = cast(glyph, FT_BitmapGlyph)

    bitmap = property(lambda self:
                      Bitmap(self._FT_BitmapGlyph.contents.bitmap))
    left = property(lambda self: self._FT_BitmapGlyph.contents.left)
    top = property(lambda self: self._FT_BitmapGlyph.contents.top)


class GlyphSlot(object):
    def __init__(self, slot):
        self._FT_GlyphSlot = slot

    def get_glyph(self):
        aglyph = FT_Glyph()
        error = FT_Get_Glyph(self._FT_GlyphSlot, byref(aglyph))
        if error:
            raise RuntimeError(hex(error))
        return Glyph(aglyph)

    bitmap = property(lambda self: Bitmap(self._FT_GlyphSlot.contents.bitmap))
    metrics = property(lambda self: self._FT_GlyphSlot.contents.metrics)
    next = property(lambda self: GlyphSlot(self._FT_GlyphSlot.contents.next))
    advance = property(lambda self: self._FT_GlyphSlot.contents.advance)
    format = property(lambda self: self._FT_GlyphSlot.contents.format)
    bitmap_top = property(lambda self: self._FT_GlyphSlot.contents.bitmap_top)
    bitmap_left = property(lambda self:
                           self._FT_GlyphSlot.contents.bitmap_left)


class Face(object):
    def __init__(self, filename, index=0):
        library = get_handle()
        face = FT_Face()
        self._FT_Face = None
        # error = FT_New_Face( library, filename, 0, byref(face) )
        u_filename = c_char_p(filename.encode('utf-8'))
        error = FT_New_Face(library, u_filename, index, byref(face))
        if error:
            raise RuntimeError(hex(error))
        self._filename = filename
        self._index = index
        self._FT_Face = face

    def __del__(self):
        if self._FT_Face is not None and FT_Done_Face is not None:
            FT_Done_Face(self._FT_Face)

    def attach_file(self, filename):
        error = FT_Attach_File(self._FT_Face, filename)
        if error:
            raise RuntimeError(hex(error))

    def set_char_size(self, width=0, height=0, hres=72, vres=72):
        error = FT_Set_Char_Size(self._FT_Face, width, height, hres, vres)
        if error:
            raise RuntimeError('Could not set size: %s' % hex(error))

    def set_pixel_sizes(self, width, height):
        error = FT_Set_Pixel_Sizes(self._FT_Face, width, height)
        if error:
            raise RuntimeError(hex(error))

    def select_charmap(self, encoding):
        error = FT_Select_Charmap(self._FT_Face, encoding)
        if error:
            raise RuntimeError(hex(error))

    def set_charmap(self, charmap):
        error = FT_Set_Charmap(self._FT_Face, charmap._FT_Charmap)
        if error:
            raise RuntimeError(hex(error))

    def get_char_index(self, charcode):
        if isinstance(charcode, string_types):
            charcode = ord(charcode)
        return FT_Get_Char_Index(self._FT_Face, charcode)

    def get_first_char(self):
        agindex = FT_UInt()
        charcode = FT_Get_First_Char(self._FT_Face, byref(agindex))
        return charcode, agindex.value

    def get_next_char(self, charcode, agindex):
        agindex = FT_UInt(0)  # agindex )
        charcode = FT_Get_Next_Char(self._FT_Face, charcode, byref(agindex))
        return charcode, agindex.value

    def get_name_index(self, name):
        return FT_Get_Name_Index(self._FT_Face, name)

    def set_transform(self, matrix, delta):
        FT_Set_Transform(self._FT_Face,
                         byref(matrix), byref(delta))

    def select_size(self, strike_index):
        error = FT_Select_Size(self._FT_Face, strike_index)
        if error:
            raise RuntimeError(hex(error))

    def load_glyph(self, index, flags=FT_LOAD_RENDER):
        error = FT_Load_Glyph(self._FT_Face, index, flags)
        if error:
            raise RuntimeError(hex(error))

    def load_char(self, char, flags=FT_LOAD_RENDER):
        if len(char) == 1:
            char = ord(char)
        error = FT_Load_Char(self._FT_Face, char, flags)
        if error:
            raise RuntimeError(hex(error))

    def get_advance(self, gindex, flags):
        padvance = FT_Fixed(0)
        error = FT_Get_Advance(self._FT_Face, gindex, flags, byref(padvance))
        if error:
            raise RuntimeError(hex(error))
        return padvance.value

    def get_kerning(self, left, right, mode=FT_KERNING_DEFAULT):
        left_glyph = self.get_char_index(left)
        right_glyph = self.get_char_index(right)
        kerning = FT_Vector(0, 0)
        error = FT_Get_Kerning(self._FT_Face,
                               left_glyph, right_glyph, mode, byref(kerning))
        if error:
            raise RuntimeError(hex(error))
        return kerning

    def get_format(self):
        return FT_Get_X11_Font_Format(self._FT_Face)

    sfnt_name_count = property(lambda self:
                               FT_Get_Sfnt_Name_Count(self._FT_Face))
    postscript_name = property(lambda self:
                               FT_Get_Postscript_Name(self._FT_Face))
    num_faces = property(lambda self: self._FT_Face.contents.num_faces)
    face_index = property(lambda self: self._FT_Face.contents.face_index)
    face_flags = property(lambda self: self._FT_Face.contents.face_flags)
    style_flags = property(lambda self: self._FT_Face.contents.style_flags)
    num_glyphs = property(lambda self: self._FT_Face.contents.num_glyphs)
    family_name = property(lambda self: self._FT_Face.contents.family_name)
    style_name = property(lambda self: self._FT_Face.contents.style_name)
    num_fixed_sizes = property(lambda self:
                               self._FT_Face.contents.num_fixed_sizes)
    num_charmaps = property(lambda self: self._FT_Face.contents.num_charmaps)
    units_per_EM = property(lambda self: self._FT_Face.contents.units_per_EM)
    ascender = property(lambda self: self._FT_Face.contents.ascender)
    descender = property(lambda self: self._FT_Face.contents.descender)
    height = property(lambda self: self._FT_Face.contents.height)
    max_advance_width = property(lambda self:
                                 self._FT_Face.contents.max_advance_width)
    max_advance_height = property(lambda self:
                                  self._FT_Face.contents.max_advance_height)
    underline_position = property(lambda self:
                                  self._FT_Face.contents.underline_position)
    underline_thickness = property(lambda self:
                                   self._FT_Face.contents.underline_thickness)
    glyph = property(lambda self: GlyphSlot(self._FT_Face.contents.glyph))



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from __future__ import division

from .node import Node


class SubScene(Node):
    """A Node subclass that serves as a marker and parent node for certain
    branches of the scenegraph.
    
    SubScene nodes are used as the top-level node for the internal scenes of
    a canvas and a view box.
    """
    def __init__(self, **kwargs):
        Node.__init__(self, **kwargs)
        self.document = self



from ..node import Node


class Anchor(Node):
    """
    Anchor is a node derives parts of its transform from some other
    coordinate system in the scene.

    The purpose is to allow children of an Anchor to draw using a position
    (and optionally rotation) specified by one coordinate system, and scaling/
    projection specified by another.

    For example, text attached to a point in a 3D scene should be drawn in
    a coordinate system with a simple relationship to the screen pixels, but
    should derive its location from a position within the 3D coordinate
    system::

        root = Box()
        view = ViewBox(parent=box)
        plot = LineVisual(parent=ViewBox)
        anchor = Anchor(parent=root, anchor_to=plot, anchor_pos=(10, 0))
        text = Text(parent=anchor,
                    text="Always points to (10,0) relative to line.")

    """



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014, Nicolas P. Rougier
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
import re
import math

from . base import units
from .. import logger


class Length(object):

    def __init__(self, content, mode='x', parent=None):

        if not content:
            self._unit = None
            self._value = 0
            self._computed_value = 0
            return

        re_number = r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
        re_unit = r'em|ex|px|in|cm|mm|pt|pc|%'
        re_length = r'(?P<value>%s)\s*(?P<unit>%s)*' % (re_number, re_unit)
        match = re.match(re_length, content)

        if match:
            self._value = float(match.group("value"))
            self._unit = match.group("unit") or "px"
        else:
            self._value = 0.0
            self._unit = None

        scale = 1
        if self._unit == '%':
            if not parent:
                logger.warn("No parent for computing length using percent")
            elif hasattr(parent, 'viewport'):
                w, h = parent.viewport
                if mode == 'x':
                    scale = w
                elif mode == 'y':
                    scale = h
                elif mode == 'xy':
                    scale = math.sqrt(w * w + h * h) / math.sqrt(2.0)
            else:
                logger.warn("Parent doesn't have a viewport")

        self._computed_value = self._value * units[self._unit] * scale

    def __float__(self):
        return self._computed_value

    @property
    def value(self):
        return self._computed_value

    def __repr__(self):
        if self._unit:
            return "%g%s" % (self._value, self._unit)
        else:
            return "%g" % (self._value)


class XLength(Length):

    def __init__(self, content, parent=None):
        Length.__init__(self, content, 'x', parent)


class YLength(Length):

    def __init__(self, content, parent=None):
        Length.__init__(self, content, 'y', parent)


class XYLength(Length):

    def __init__(self, content, parent=None):
        Length.__init__(self, content, 'xy', parent)



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from __future__ import division

import numpy as np

from .widget import Widget
from ..subscene import SubScene
from ..cameras import make_camera, BaseCamera
from ...ext.six import string_types
from ...visuals.filters import Clipper


class ViewBox(Widget):
    """ Provides a rectangular widget to which its subscene is rendered.

    Three classes work together when using a ViewBox:
    * The :class:`SubScene` class describes a "world" coordinate system and the
    entities that live inside it.
    * ViewBox is a "window" through which we view the
    subscene. Multiple ViewBoxes may view the same subscene.
    * :class:`Camera` describes both the perspective from which the
    subscene is rendered, and the way user interaction affects that
    perspective.

    In general it is only necessary to create the ViewBox; a SubScene and
    Camera will be generated automatically.

    Parameters
    ----------
    camera : instance of Camera | str | None
        The camera through which to view the SubScene. If None, then a
        PanZoomCamera (2D interaction) is used. If str, then the string is
        used as the argument to :func:`make_camera`.
    **kwargs : dict
        Extra keyword arguments to pass to `Widget`.
    """
    def __init__(self, camera=None, **kwargs):
        self._camera = None
        self._scene = None
        Widget.__init__(self, **kwargs)
        self.interactive = True

        # Each viewbox has an internal scene node, which has a transform that
        # represents the transformation imposed by camera.
        if self.name is not None:
            name = str(self.name) + "_Scene"
        else:
            name = None
            
        self._scene = SubScene(name=name, parent=self)
        self._scene._clipper = Clipper()
        self._scene.clip_children = True
        self.transforms.changed.connect(self._update_scene_clipper)
        
        # Camera is a helper object that handles scene transformation
        # and user interaction.
        if camera is None:
            camera = 'base'
        if isinstance(camera, string_types):
            self.camera = make_camera(camera, parent=self.scene)
        elif isinstance(camera, BaseCamera):
            self.camera = camera
        else:
            raise TypeError('Argument "camera" must be None, str, or Camera.')

    @property
    def camera(self):
        """ Get/set the Camera in use by this ViewBox

        If a string is given (e.g. 'panzoom', 'turntable', 'fly'). A
        corresponding camera is selected if it already exists in the
        scene, otherwise a new camera is created.

        The camera object is made a child of the scene (if it is not
        already in the scene).

        Multiple cameras can exist in one scene, although only one can
        be active at a time. A single camera can be used by multiple
        viewboxes at the same time.
        """
        return self._camera

    @camera.setter
    def camera(self, cam):
        if isinstance(cam, string_types):
            # Try to select an existing camera
            for child in self.scene.children:
                if isinstance(child, BaseCamera):
                    this_cam_type = child.__class__.__name__.lower()[:-6]
                    if this_cam_type == cam:
                        self.camera = child
                        return
            else:
                # No such camera yet, create it then
                self.camera = make_camera(cam)
            
        elif isinstance(cam, BaseCamera):
            # Ensure that the camera is in the scene
            if not self.is_in_scene(cam):
                cam.parent = self.scene
            # Disconnect / connect
            if self._camera is not None:
                self._camera._viewbox_unset(self)
            self._camera = cam
            if self._camera is not None:
                self._camera._viewbox_set(self)
            # Update view
            cam.view_changed()
        
        else:
            raise ValueError('Not a camera object.')

    def is_in_scene(self, node):
        """Get whether the given node is inside the scene of this viewbox.

        Parameters
        ----------
        node : instance of Node
            The node.
        """
        return self.scene.is_child(node)
    
    def get_scene_bounds(self, dim=None):
        """Get the total bounds based on the visuals present in the scene

        Parameters
        ----------
        dim : int | None
            Dimension to return.

        Returns
        -------
        bounds : list | tuple
            If ``dim is None``, Returns a list of 3 tuples, otherwise
            the bounds for the requested dimension.
        """
        # todo: handle sub-children
        # todo: handle transformations
        # Init
        bounds = [(np.inf, -np.inf), (np.inf, -np.inf), (np.inf, -np.inf)]
        # Get bounds of all children
        for ob in self.scene.children:
            if hasattr(ob, 'bounds'):
                for axis in (0, 1, 2):
                    if (dim is not None) and dim != axis:
                        continue
                    b = ob.bounds(axis)
                    if b is not None:
                        b = min(b), max(b)  # Ensure correct order
                        bounds[axis] = (min(bounds[axis][0], b[0]), 
                                        max(bounds[axis][1], b[1]))
        # Set defaults
        for axis in (0, 1, 2):
            if any(np.isinf(bounds[axis])):
                bounds[axis] = -1, 1
        
        if dim is not None:
            return bounds[dim]
        else:
            return bounds
    
    @property
    def scene(self):
        """ The root node of the scene viewed by this ViewBox.
        """
        return self._scene

    def add(self, node):
        """ Add an Node to the scene for this ViewBox.

        This is a convenience method equivalent to
        `node.parent = viewbox.scene`

        Parameters
        ----------
        node : instance of Node
            The node to add.
        """
        node.parent = self.scene

    def on_resize(self, event):
        """Resize event handler

        Parameters
        ----------
        event : instance of Event
            The event.
        """
        if self._scene is None:
            # happens during init
            return
        self._update_scene_clipper()
        
    def _update_scene_clipper(self, event=None):
        tr = self.get_transform('visual', 'framebuffer')
        self._scene._clipper.bounds = tr.map(self.inner_rect)



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""
Provides classes representing different transform types suitable for
use with visuals and scenes.
"""


from .base_transform import BaseTransform  # noqa
from .linear import (NullTransform, STTransform,  # noqa
                     MatrixTransform,  MatrixTransform)  # noqa
from .nonlinear import LogTransform, PolarTransform  # noqa
from .interactive import PanZoomTransform
from .chain import ChainTransform  # noqa
from ._util import arg_to_array, arg_to_vec4, as_vec4, TransformCache  # noqa
from .transform_system import TransformSystem

__all__ = ['NullTransform', 'STTransform', 'MatrixTransform',
           'MatrixTransform', 'LogTransform', 'PolarTransform',
           'ChainTransform', 'TransformSystem', 'PanZoomTransform']

transform_types = {}
for o in list(globals().values()):
    try:
        if issubclass(o, BaseTransform) and o is not BaseTransform:
            name = o.__name__[:-len('Transform')].lower()
            transform_types[name] = o
    except TypeError:
        continue


def create_transform(type, *args, **kwargs):
    return transform_types[type](*args, **kwargs)



from __future__ import print_function, unicode_literals, absolute_import, division


class InternalError(Exception):
    pass


class ConstraintNotFound(Exception):
    pass


class RequiredFailure(Exception):
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""
Line visual implementing Agg- and GL-based drawing modes.
"""

from __future__ import division

import numpy as np

from ... import gloo, glsl
from ...color import Color, ColorArray, get_colormap
from ...ext.six import string_types
from ..shaders import Function
from ..visual import Visual, CompoundVisual
from ...util.profiler import Profiler

from .dash_atlas import DashAtlas


vec2to4 = Function("""
    vec4 vec2to4(vec2 inp) {
        return vec4(inp, 0, 1);
    }
""")

vec3to4 = Function("""
    vec4 vec3to4(vec3 inp) {
        return vec4(inp, 1);
    }
""")


"""
TODO:

* Agg support is very minimal; needs attention.
* Optimization--avoid creating new buffers, avoid triggering program
  recompile.
"""


joins = {'miter': 0, 'round': 1, 'bevel': 2}

caps = {'': 0, 'none': 0, '.': 0,
        'round': 1, ')': 1, '(': 1, 'o': 1,
        'triangle in': 2, '<': 2,
        'triangle out': 3, '>': 3,
        'square': 4, '=': 4, 'butt': 4,
        '|': 5}


class LineVisual(CompoundVisual):
    """Line visual

    Parameters
    ----------
    pos : array
        Array of shape (..., 2) or (..., 3) specifying vertex coordinates.
    color : Color, tuple, or array
        The color to use when drawing the line. If an array is given, it
        must be of shape (..., 4) and provide one rgba color per vertex.
        Can also be a colormap name, or appropriate `Function`.
    width:
        The width of the line in px. Line widths > 1px are only
        guaranteed to work when using 'agg' method.
    connect : str or array
        Determines which vertices are connected by lines.

            * "strip" causes the line to be drawn with each vertex
              connected to the next.
            * "segments" causes each pair of vertices to draw an
              independent line segment
            * numpy arrays specify the exact set of segment pairs to
              connect.

    method : str
        Mode to use for drawing.

            * "agg" uses anti-grain geometry to draw nicely antialiased lines
              with proper joins and endcaps.
            * "gl" uses OpenGL's built-in line rendering. This is much faster,
              but produces much lower-quality results and is not guaranteed to
              obey the requested line width or join/endcap styles.

    antialias : bool
        Enables or disables antialiasing.
        For method='gl', this specifies whether to use GL's line smoothing,
        which may be unavailable or inconsistent on some platforms.
    """
    def __init__(self, pos=None, color=(0.5, 0.5, 0.5, 1), width=1,
                 connect='strip', method='gl', antialias=False):
        self._line_visual = None

        self._changed = {'pos': False, 'color': False, 'width': False,
                         'connect': False}

        self._pos = None
        self._color = None
        self._width = None
        self._connect = None
        self._bounds = None
        self._antialias = None
        self._method = 'none'

        CompoundVisual.__init__(self, [])

        # don't call subclass set_data; these often have different
        # signatures.
        LineVisual.set_data(self, pos=pos, color=color, width=width,
                            connect=connect)
        self.antialias = antialias
        self.method = method

    @property
    def antialias(self):
        return self._antialias

    @antialias.setter
    def antialias(self, aa):
        self._antialias = bool(aa)
        self.update()

    @property
    def method(self):
        """The current drawing method"""
        return self._method

    @method.setter
    def method(self, method):
        if method not in ('agg', 'gl'):
            raise ValueError('method argument must be "agg" or "gl".')
        if method == self._method:
            return

        self._method = method
        if self._line_visual is not None:
            self.remove_subvisual(self._line_visual)

        if method == 'gl':
            self._line_visual = _GLLineVisual(self)
        elif method == 'agg':
            self._line_visual = _AggLineVisual(self)
        self.add_subvisual(self._line_visual)

        for k in self._changed:
            self._changed[k] = True

    def set_data(self, pos=None, color=None, width=None, connect=None):
        """ Set the data used to draw this visual.

        Parameters
        ----------
        pos : array
            Array of shape (..., 2) or (..., 3) specifying vertex coordinates.
        color : Color, tuple, or array
            The color to use when drawing the line. If an array is given, it
            must be of shape (..., 4) and provide one rgba color per vertex.
        width:
            The width of the line in px. Line widths < 1 px will be rounded up
            to 1 px when using the 'gl' method.
        connect : str or array
            Determines which vertices are connected by lines.
            * "strip" causes the line to be drawn with each vertex
              connected to the next.
            * "segments" causes each pair of vertices to draw an
              independent line segment
            * int numpy arrays specify the exact set of segment pairs to
              connect.
            * bool numpy arrays specify which _adjacent_ pairs to connect.
        """
        if pos is not None:
            self._bounds = None
            self._pos = pos
            self._changed['pos'] = True

        if color is not None:
            self._color = color
            self._changed['color'] = True

        if width is not None:
            self._width = width
            self._changed['width'] = True

        if connect is not None:
            self._connect = connect
            self._changed['connect'] = True

        self.update()

    @property
    def color(self):
        return self._color

    @property
    def width(self):
        return self._width

    @property
    def connect(self):
        return self._connect

    @property
    def pos(self):
        return self._pos

    def _interpret_connect(self):
        if isinstance(self._connect, np.ndarray):
            # Convert a boolean connection array to a vertex index array
            if self._connect.ndim == 1 and self._connect.dtype == bool:
                index = np.empty((len(self._connect), 2), dtype=np.uint32)
                index[:] = np.arange(len(self._connect))[:, np.newaxis]
                index[:, 1] += 1
                return index[self._connect]
            elif self._connect.ndim == 2 and self._connect.shape[1] == 2:
                return self._connect.astype(np.uint32)
            else:
                raise TypeError("Got invalid connect array of shape %r and "
                                "dtype %r" % (self._connect.shape,
                                              self._connect.dtype))
        else:
            return self._connect

    def _interpret_color(self):
        if isinstance(self._color, string_types):
            try:
                colormap = get_colormap(self._color)
                color = Function(colormap.glsl_map)
            except KeyError:
                color = Color(self._color).rgba
        elif isinstance(self._color, Function):
            color = Function(self._color)
        else:
            color = ColorArray(self._color).rgba
            if len(color) == 1:
                color = color[0]
        return color

    def _compute_bounds(self, axis, view):
        """Get the bounds

        Parameters
        ----------
        mode : str
            Describes the type of boundary requested. Can be "visual", "data",
            or "mouse".
        axis : 0, 1, 2
            The axis along which to measure the bounding values, in
            x-y-z order.
        """
        # Can and should we calculate bounds?
        if (self._bounds is None) and self._pos is not None:
            pos = self._pos
            self._bounds = [(pos[:, d].min(), pos[:, d].max())
                            for d in range(pos.shape[1])]
        # Return what we can
        if self._bounds is None:
            return
        else:
            if axis < len(self._bounds):
                return self._bounds[axis]
            else:
                return (0, 0)

    def _prepare_draw(self, view):
        if self._width == 0:
            return False
        CompoundVisual._prepare_draw(self, view)


class _GLLineVisual(Visual):
    VERTEX_SHADER = """
        varying vec4 v_color;

        void main(void) {
            gl_Position = $transform($to_vec4($position));
            v_color = $color;
        }
    """

    FRAGMENT_SHADER = """
        varying vec4 v_color;
        void main() {
            gl_FragColor = v_color;
        }
    """

    def __init__(self, parent):
        self._parent = parent
        self._pos_vbo = gloo.VertexBuffer()
        self._color_vbo = gloo.VertexBuffer()
        self._connect_ibo = gloo.IndexBuffer()
        self._connect = None

        Visual.__init__(self, vcode=self.VERTEX_SHADER,
                        fcode=self.FRAGMENT_SHADER)
        self.set_gl_state('translucent')

    def _prepare_transforms(self, view):
        xform = view.transforms.get_transform()
        view.view_program.vert['transform'] = xform

    def _prepare_draw(self, view):
        prof = Profiler()

        if self._parent._changed['pos']:
            if self._parent._pos is None:
                return False
            # todo: does this result in unnecessary copies?
            pos = np.ascontiguousarray(self._parent._pos.astype(np.float32))
            self._pos_vbo.set_data(pos)
            self._program.vert['position'] = self._pos_vbo
            if pos.shape[-1] == 2:
                self._program.vert['to_vec4'] = vec2to4
            elif pos.shape[-1] == 3:
                self._program.vert['to_vec4'] = vec3to4
            else:
                raise TypeError("Got bad position array shape: %r"
                                % (pos.shape,))

        if self._parent._changed['color']:
            color = self._parent._interpret_color()
            # If color is not visible, just quit now
            if isinstance(color, Color) and color.is_blank:
                return False
            if isinstance(color, Function):
                # TODO: Change to the parametric coordinate once that is done
                self._program.vert['color'] = color(
                    '(gl_Position.x + 1.0) / 2.0')
            else:
                if color.ndim == 1:
                    self._program.vert['color'] = color
                else:
                    self._color_vbo.set_data(color)
                    self._program.vert['color'] = self._color_vbo

        # Do we want to use OpenGL, and can we?
        GL = None
        from ...app._default_app import default_app
        if default_app is not None and \
                default_app.backend_name != 'ipynb_webgl':
            try:
                import OpenGL.GL as GL
            except Exception:  # can be other than ImportError sometimes
                pass

        # Turn on line smooth and/or line width
        try:
            if GL:
                if self._parent._antialias:
                    GL.glEnable(GL.GL_LINE_SMOOTH)
                else:
                    GL.glDisable(GL.GL_LINE_SMOOTH)
                px_scale = self.transforms.pixel_scale
                width = px_scale * self._parent._width
                GL.glLineWidth(max(width, 1.))
        except Exception:
            # FIXME: on AppVeyor, the above code doesn't run properly, possibly
            #        because some of these OpenGL methods have been deprecated
            pass

        if self._parent._changed['connect']:
            self._connect = self._parent._interpret_connect()
            if isinstance(self._connect, np.ndarray):
                self._connect_ibo.set_data(self._connect)
        if self._connect is None:
            return False

        prof('prepare')

        # Draw
        if isinstance(self._connect, string_types) and \
                self._connect == 'strip':
            self._draw_mode = 'line_strip'
            self._index_buffer = None
        elif isinstance(self._connect, string_types) and \
                self._connect == 'segments':
            self._draw_mode = 'lines'
            self._index_buffer = None
        elif isinstance(self._connect, np.ndarray):
            self._draw_mode = 'lines'
            self._index_buffer = self._connect_ibo
        else:
            raise ValueError("Invalid line connect mode: %r" % self._connect)

        prof('draw')


class _AggLineVisual(Visual):
    _agg_vtype = np.dtype([('a_position', np.float32, 2),
                           ('a_tangents', np.float32, 4),
                           ('a_segment',  np.float32, 2),
                           ('a_angles',   np.float32, 2),
                           ('a_texcoord', np.float32, 2),
                           ('alength', np.float32, 1),
                           ('color', np.float32, 4)])

    VERTEX_SHADER = glsl.get('lines/agg.vert')
    FRAGMENT_SHADER = glsl.get('lines/agg.frag')

    def __init__(self, parent):
        self._parent = parent
        self._vbo = gloo.VertexBuffer()

        self._pos = None
        self._color = None

        self._da = DashAtlas()
        dash_index, dash_period = self._da['solid']
        self._U = dict(dash_index=dash_index, dash_period=dash_period,
                       linejoin=joins['round'],
                       linecaps=(caps['round'], caps['round']),
                       dash_caps=(caps['round'], caps['round']),
                       antialias=1.0)
        self._dash_atlas = gloo.Texture2D(self._da._data)

        Visual.__init__(self, vcode=self.VERTEX_SHADER,
                        fcode=self.FRAGMENT_SHADER)
        self._index_buffer = gloo.IndexBuffer()
        self.set_gl_state('translucent', depth_test=False)
        self._draw_mode = 'triangles'

    def _prepare_transforms(self, view):
        data_doc = view.get_transform('visual', 'document')
        doc_px = view.get_transform('document', 'framebuffer')
        px_ndc = view.get_transform('framebuffer', 'render')

        vert = view.view_program.vert
        vert['transform'] = data_doc
        vert['doc_px_transform'] = doc_px
        vert['px_ndc_transform'] = px_ndc

    def _prepare_draw(self, view):
        bake = False
        if self._parent._changed['pos']:
            if self._parent._pos is None:
                return False
            # todo: does this result in unnecessary copies?
            self._pos = np.ascontiguousarray(
                self._parent._pos.astype(np.float32))
            bake = True

        if self._parent._changed['color']:
            self._color = self._parent._interpret_color()
            bake = True

        if self._parent._changed['connect']:
            if self._parent._connect not in [None, 'strip']:
                raise NotImplementedError("Only 'strip' connection mode "
                                          "allowed for agg-method lines.")

        if bake:
            V, I = self._agg_bake(self._pos, self._color)
            self._vbo.set_data(V)
            self._index_buffer.set_data(I)

        #self._program.prepare()
        self.shared_program.bind(self._vbo)
        uniforms = dict(closed=False, miter_limit=4.0, dash_phase=0.0,
                        linewidth=self._parent._width)
        for n, v in uniforms.items():
            self.shared_program[n] = v
        for n, v in self._U.items():
            self.shared_program[n] = v
        self.shared_program['u_dash_atlas'] = self._dash_atlas

    @classmethod
    def _agg_bake(cls, vertices, color, closed=False):
        """
        Bake a list of 2D vertices for rendering them as thick line. Each line
        segment must have its own vertices because of antialias (this means no
        vertex sharing between two adjacent line segments).
        """

        n = len(vertices)
        P = np.array(vertices).reshape(n, 2).astype(float)
        idx = np.arange(n)  # used to eventually tile the color array

        dx, dy = P[0] - P[-1]
        d = np.sqrt(dx*dx+dy*dy)

        # If closed, make sure first vertex = last vertex (+/- epsilon=1e-10)
        if closed and d > 1e-10:
            P = np.append(P, P[0]).reshape(n+1, 2)
            idx = np.append(idx, idx[-1])
            n += 1

        V = np.zeros(len(P), dtype=cls._agg_vtype)
        V['a_position'] = P

        # Tangents & norms
        T = P[1:] - P[:-1]

        N = np.sqrt(T[:, 0]**2 + T[:, 1]**2)
        # T /= N.reshape(len(T),1)
        V['a_tangents'][+1:, :2] = T
        V['a_tangents'][0, :2] = T[-1] if closed else T[0]
        V['a_tangents'][:-1, 2:] = T
        V['a_tangents'][-1, 2:] = T[0] if closed else T[-1]

        # Angles
        T1 = V['a_tangents'][:, :2]
        T2 = V['a_tangents'][:, 2:]
        A = np.arctan2(T1[:, 0]*T2[:, 1]-T1[:, 1]*T2[:, 0],
                       T1[:, 0]*T2[:, 0]+T1[:, 1]*T2[:, 1])
        V['a_angles'][:-1, 0] = A[:-1]
        V['a_angles'][:-1, 1] = A[+1:]

        # Segment
        L = np.cumsum(N)
        V['a_segment'][+1:, 0] = L
        V['a_segment'][:-1, 1] = L
        # V['a_lengths'][:,2] = L[-1]

        # Step 1: A -- B -- C  =>  A -- B, B' -- C
        V = np.repeat(V, 2, axis=0)[1:-1]
        V['a_segment'][1:] = V['a_segment'][:-1]
        V['a_angles'][1:] = V['a_angles'][:-1]
        V['a_texcoord'][0::2] = -1
        V['a_texcoord'][1::2] = +1
        idx = np.repeat(idx, 2)[1:-1]

        # Step 2: A -- B, B' -- C  -> A0/A1 -- B0/B1, B'0/B'1 -- C0/C1
        V = np.repeat(V, 2, axis=0)
        V['a_texcoord'][0::2, 1] = -1
        V['a_texcoord'][1::2, 1] = +1
        idx = np.repeat(idx, 2)

        I = np.resize(np.array([0, 1, 2, 1, 2, 3], dtype=np.uint32),
                      (n-1)*(2*3))
        I += np.repeat(4*np.arange(n-1, dtype=np.uint32), 6)

        # Length
        V['alength'] = L[-1] * np.ones(len(V))

        # Color
        if color.ndim == 1:
            color = np.tile(color, (len(V), 1))
        elif color.ndim == 2 and len(color) == n:
            color = color[idx]
        else:
            raise ValueError('Color length %s does not match number of '
                             'vertices %s' % (len(color), n))
        V['color'] = color

        return V, I



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from __future__ import division  # just to be safe...

import numpy as np

from ..ext.six import string_types


###############################################################################
# Utility functions
def _check_color_dim(val):
    """Ensure val is Nx(n_col), usually Nx3"""
    val = np.atleast_2d(val)
    if val.shape[1] not in (3, 4):
        raise RuntimeError('Value must have second dimension of size 3 or 4')
    return val, val.shape[1]


###############################################################################
# RGB<->HEX conversion

def _hex_to_rgba(hexs):
    """Convert hex to rgba, permitting alpha values in hex"""
    hexs = np.atleast_1d(np.array(hexs, '|U9'))
    out = np.ones((len(hexs), 4), np.float32)
    for hi, h in enumerate(hexs):
        assert isinstance(h, string_types)
        off = 1 if h[0] == '#' else 0
        assert len(h) in (6+off, 8+off)
        e = (len(h)-off) // 2
        out[hi, :e] = [int(h[i:i+2], 16) / 255.
                       for i in range(off, len(h), 2)]
    return out


def _rgb_to_hex(rgbs):
    """Convert rgb to hex triplet"""
    rgbs, n_dim = _check_color_dim(rgbs)
    return np.array(['#%02x%02x%02x' % tuple((255*rgb[:3]).astype(np.uint8))
                     for rgb in rgbs], '|U7')


###############################################################################
# RGB<->HSV conversion

def _rgb_to_hsv(rgbs):
    """Convert Nx3 or Nx4 rgb to hsv"""
    rgbs, n_dim = _check_color_dim(rgbs)
    hsvs = list()
    for rgb in rgbs:
        rgb = rgb[:3]  # don't use alpha here
        idx = np.argmax(rgb)
        val = rgb[idx]
        c = val - np.min(rgb)
        if c == 0:
            hue = 0
            sat = 0
        else:
            if idx == 0:  # R == max
                hue = ((rgb[1] - rgb[2]) / c) % 6
            elif idx == 1:  # G == max
                hue = (rgb[2] - rgb[0]) / c + 2
            else:  # B == max
                hue = (rgb[0] - rgb[1]) / c + 4
            hue *= 60
            sat = c / val
        hsv = [hue, sat, val]
        hsvs.append(hsv)
    hsvs = np.array(hsvs, dtype=np.float32)
    if n_dim == 4:
        hsvs = np.concatenate((hsvs, rgbs[:, 3]), axis=1)
    return hsvs


def _hsv_to_rgb(hsvs):
    """Convert Nx3 or Nx4 hsv to rgb"""
    hsvs, n_dim = _check_color_dim(hsvs)
    # In principle, we *might* be able to vectorize this, but might as well
    # wait until a compelling use case appears
    rgbs = list()
    for hsv in hsvs:
        c = hsv[1] * hsv[2]
        m = hsv[2] - c
        hp = hsv[0] / 60
        x = c * (1 - abs(hp % 2 - 1))
        if 0 <= hp < 1:
            r, g, b = c, x, 0
        elif hp < 2:
            r, g, b = x, c, 0
        elif hp < 3:
            r, g, b = 0, c, x
        elif hp < 4:
            r, g, b = 0, x, c
        elif hp < 5:
            r, g, b = x, 0, c
        else:
            r, g, b = c, 0, x
        rgb = [r + m, g + m, b + m]
        rgbs.append(rgb)
    rgbs = np.array(rgbs, dtype=np.float32)
    if n_dim == 4:
        rgbs = np.concatenate((rgbs, hsvs[:, 3]), axis=1)
    return rgbs


###############################################################################
# RGB<->CIELab conversion

# These numbers are adapted from MIT-licensed MATLAB code for
# Lab<->RGB conversion. They provide an XYZ<->RGB conversion matrices,
# w/D65 white point normalization built in.

#_rgb2xyz = np.array([[0.412453, 0.357580, 0.180423],
#                     [0.212671, 0.715160, 0.072169],
#                     [0.019334, 0.119193, 0.950227]])
#_white_norm = np.array([0.950456, 1.0, 1.088754])
#_rgb2xyz /= _white_norm[:, np.newaxis]
#_rgb2xyz_norm = _rgb2xyz.T
_rgb2xyz_norm = np.array([[0.43395276, 0.212671, 0.01775791],
                         [0.37621941, 0.71516, 0.10947652],
                         [0.18982783, 0.072169, 0.87276557]])

#_xyz2rgb = np.array([[3.240479, -1.537150, -0.498535],
#                     [-0.969256, 1.875992, 0.041556],
#                     [0.055648, -0.204043, 1.057311]])
#_white_norm = np.array([0.950456, 1., 1.088754])
#_xyz2rgb *= _white_norm[np.newaxis, :]
_xyz2rgb_norm = np.array([[3.07993271, -1.53715, -0.54278198],
                          [-0.92123518, 1.875992, 0.04524426],
                          [0.05289098, -0.204043, 1.15115158]])


def _rgb_to_lab(rgbs):
    rgbs, n_dim = _check_color_dim(rgbs)
    # convert RGB->XYZ
    xyz = rgbs[:, :3].copy()  # a misnomer for now but will end up being XYZ
    over = xyz > 0.04045
    xyz[over] = ((xyz[over] + 0.055) / 1.055) ** 2.4
    xyz[~over] /= 12.92
    xyz = np.dot(xyz, _rgb2xyz_norm)
    over = xyz > 0.008856
    xyz[over] = xyz[over] ** (1. / 3.)
    xyz[~over] = 7.787 * xyz[~over] + 0.13793103448275862

    # Convert XYZ->LAB
    L = (116. * xyz[:, 1]) - 16
    a = 500 * (xyz[:, 0] - xyz[:, 1])
    b = 200 * (xyz[:, 1] - xyz[:, 2])
    labs = [L, a, b]
    # Append alpha if necessary
    if n_dim == 4:
        labs.append(np.atleast1d(rgbs[:, 3]))
    labs = np.array(labs, order='F').T  # Becomes 'C' order b/c of .T
    return labs


def _lab_to_rgb(labs):
    """Convert Nx3 or Nx4 lab to rgb"""
    # adapted from BSD-licensed work in MATLAB by Mark Ruzon
    # Based on ITU-R Recommendation BT.709 using the D65
    labs, n_dim = _check_color_dim(labs)

    # Convert Lab->XYZ (silly indexing used to preserve dimensionality)
    y = (labs[:, 0] + 16.) / 116.
    x = (labs[:, 1] / 500.) + y
    z = y - (labs[:, 2] / 200.)
    xyz = np.concatenate(([x], [y], [z]))  # 3xN
    over = xyz > 0.2068966
    xyz[over] = xyz[over] ** 3.
    xyz[~over] = (xyz[~over] - 0.13793103448275862) / 7.787

    # Convert XYZ->LAB
    rgbs = np.dot(_xyz2rgb_norm, xyz).T
    over = rgbs > 0.0031308
    rgbs[over] = 1.055 * (rgbs[over] ** (1. / 2.4)) - 0.055
    rgbs[~over] *= 12.92
    if n_dim == 4:
        rgbs = np.concatenate((rgbs, labs[:, 3]), axis=1)
    rgbs = np.clip(rgbs, 0., 1.)
    return rgbs
